/*
   vdev: a virtual device manager for *nix
   Copyright (C) 2014  Jude Nelson

   This program is dual-licensed: you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 or later as 
   published by the Free Software Foundation. For the terms of this 
   license, see LICENSE.GPLv3+ or <http://www.gnu.org/licenses/>.

   You are free to use this program under the terms of the GNU General
   Public License, but WITHOUT ANY WARRANTY; without even the implied 
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   Alternatively, you are free to use this program under the terms of the 
   Internet Software Consortium License, but WITHOUT ANY WARRANTY; without
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   For the terms of this license, see LICENSE.ISC or 
   <http://www.isc.org/downloads/software-support-policy/isc-license/>.
*/
 
#include "vdev-modprobe-util.h"

#include <fcntl.h>     // O_RDONLY
#include <sys/stat.h>  // open()
#include <assert.h>
#include <signal.h>

#include <fcntl.h>     // O_RDONLY
#include <sys/stat.h>  // open()
#include <assert.h>

// Usage: strncpy_t(str1, sizeof(str1), str2, strlen(str2));
// Copy string "in" with at most "insz" chars to buffer "out", which
// is "outsz" bytes long. The output is always 0-terminated. Unlike
// strncpy(), strncpy_t() does not zero fill remaining space in the
// output buffer:
char* strncpy_t(char* out, size_t outsz, const char* in, size_t insz)
{
   assert(outsz > 0);
   while(--outsz > 0 && insz > 0 && *in) { *out++ = *in++; insz--; }
   *out = 0;
   return out;
}

/**
 * \brief A subprogram like popen() but with diferences explained below
 * The pipe is to read in the current process and connected to stderr in the
 * new one. The intent is to report error messages. Plus pid is returned.
 * Disclaimer: The caller must close the pipe after use
 */
FILE *epopen(const char *cmd,  pid_t *pid)
{
  int pipefd[2];
  pid_t p_id;
  FILE *pf;
  int rc;

  // Create the pipe
  if(pipe(pipefd)) {
	  
      *pid = -1;
      return NULL;
  }

  p_id = fork();
 
  switch(p_id) {
    
    case 0: /* child */
      close(pipefd[0]);
      close(STDERR_FILENO);
      dup2(pipefd[1], STDERR_FILENO);
      close(pipefd[1]);
      execl( "/bin/sh", "sh", "-c", cmd, (char *)NULL );
      fprintf( stderr, "error invoking /bin/sh: %s\n", strerror(errno) );
      exit(EXIT_FAILURE); /* reminder: we are the child */
    
    case -1: /* fork() failed ! */
      rc = errno;
      close(pipefd[0]);
      close(pipefd[1]);
      errno = rc;
      *pid = -1;
      return NULL;
    
    default: /* the parent */
      close(pipefd[1]);
      pf = fdopen(pipefd[0], "r");
      *pid = p_id;
      return pf;
  }
}

// join two paths, writing the result to dest if dest is not NULL.
// otherwise, allocate and return a buffer containing the joined paths.
// return NULL on OOM 
char* fullpath( char const* root, char const* path, char* dest ) {

   char delim = 0;
   int path_off = 0;
   int len = 0;
   
   if( root == NULL || path == NULL ) {
      return NULL;
   }
   
   len = strlen(path) + strlen(root) + 2;

   if( strlen(root) > 0 ) {
      size_t root_delim_off = strlen(root) - 1;
      if( root[root_delim_off] != '/' && path[0] != '/' ) {
         len++;
         delim = '/';
      }
      else if( root[root_delim_off] == '/' && path[0] == '/' ) {
         path_off = 1;
      }
   }

   if( dest == NULL ) {
      dest = VDEV_CALLOC( char, len );
      if( dest == NULL ) {
         return NULL;
      }
   }

   memset(dest, 0, len);

   strcpy( dest, root );
   if( delim != 0 ) {
      dest[strlen(dest)] = '/';
   }
   strcat( dest, path + path_off );

   return dest;
}

// read, but mask EINTR
// return number of bytes read on success 
// return -errno on I/O error
ssize_t modprobe_read_uninterrupted( int fd, char* buf, size_t len ) {
   
   ssize_t num_read = 0;
   
   if( buf == NULL ) {
      return -EINVAL;
   }
   
   while( (unsigned)num_read < len ) {
      ssize_t nr = read( fd, buf + num_read, len - num_read );
      if( nr < 0 ) {
         
         int errsv = -errno;
         if( errsv == -EINTR ) {
            continue;
         }
         
         return errsv;
      }
      if( nr == 0 ) {
         break;
      }
      
      num_read += nr;
   }
   
   return num_read;
}

// read a whole file, masking EINTR 
// return 0 on success, filling buf with up to len bytes 
// return -errno on open or read errors
int modprobe_read_file( char const* path, char* buf, size_t len ) {
   
   int fd = 0;
   int rc = 0;
   
   if( path == NULL || buf == NULL ) {
      return -EINVAL;
   }
   
   fd = open( path, O_RDONLY );
   if( fd < 0 ) {
      
      rc = -errno;
      fprintf(stderr, "open('%s') rc = %d\n", path, rc );
      return rc;
   }
   
   rc = modprobe_read_uninterrupted( fd, buf, len );
   if( rc < 0 ) {
      
      rc = -errno;
      fprintf(stderr, "modprobe_read_uninterrupted('%s') rc = %d\n", path, rc );
      
      close( fd );
      return rc;
   }
   
   close( fd );
   return 0;
}

// free a list of dirents 
// always succeeds
static void vdev_dirents_free( struct dirent** dirents, int num_entries ) {
   
   if( dirents == NULL ) {
      return;
   }
   
   for( int i = 0; i < num_entries; i++ ) {
      
      if( dirents[i] == NULL ) {
         continue;
      }
      
      free( dirents[i] );
      dirents[i] = NULL;
   }
   
   free( dirents );
}


// process all files in a directory with a loader, in alphanumeric order
// return 0 on success
// return -EINVAL if dir_path or loader are NULL 
// return -ENOMEM if OOM 
// return -errno on I/O error 
// return non-zero if loader fails
int vdev_load_all( char const* dir_path, vdev_dirent_loader_t loader, void* cls ) {
   
   struct dirent** dirents = NULL;
   int num_entries = 0;
   int rc = 0;
   
   if( dir_path == NULL || loader == NULL ) {
      return -EINVAL;
   }
     
   num_entries = scandir( dir_path, &dirents, NULL, alphasort );
   if( num_entries < 0 ) {
      
      rc = -errno;
      fprintf(stderr, "scandir('%s') rc = %d\n", dir_path, rc );
      return rc;
   }
   
   for( int i = 0; i < num_entries; i++ ) {
      
      // full path...
      char* fp = fullpath( dir_path, dirents[i]->d_name, NULL );
      if( fp == NULL ) {
         
         vdev_dirents_free( dirents, num_entries );
         return -ENOMEM;
      }
      
      // process this 
      rc = (*loader)( fp, cls );
      if( rc != 0 ) {
         
         fprintf(stderr, "loader('%s') rc = %d\n", fp, rc );
         
         free( fp );
         vdev_dirents_free( dirents, num_entries );
         return rc;
      }
      
      free( fp );
   }
   
   vdev_dirents_free( dirents, num_entries );
   
   return 0;
}

/**
 * \brief This function will daemonize this app
 */
void daemonize()
{
   int ret;
   pid_t pid = 0;
   int fd;

   /* Fork off the parent process.
    * The first fork will change our pid
    * but the sid and pgid will be the
    * calling process
    */
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Fork off for the second time.
    * The magical double fork. We're the session
    * leader from the code above. Since only the
    * session leader can take control of a tty
    * we will fork and exit the session leader.
    * Once the fork is done below and we use
    * the child process we will ensure we're
    * not the session leader, thus, we cannot
    * take control of a tty.
    */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }

   /* On success: The child process becomes session leader */
   if (setsid() < 0) {
      exit(EXIT_FAILURE);
   }

   /* Ignore signal sent from child to parent process */
   signal(SIGCHLD, SIG_IGN);

   /* Fork off for the second time*/
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Success: Let the parent terminate */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }
   
   /* Set new file permissions */
   umask(0);

   /* Change the working directory to /tmp */
   /* or another appropriated directory */
   ret = chdir("/tmp");
   if(ret != 0) exit(EXIT_FAILURE);

   /* Close all open file descriptors */
   for (fd = sysconf(_SC_OPEN_MAX); fd > 0; fd--) {
      close(fd);
   }

   /* Reopen stdin (fd = 0), stdout (fd = 1), stderr (fd = 2) */
   stdin  = fopen("/dev/null", "r");
   stdout = fopen("/dev/null", "w+");
   stderr = fopen("/dev/null", "w+");
}

// get the parent device of a given device, using sysfs.
// not all devices have parents; those that do will have a uevent file.
// walk up the path until we reach /
// return 0 on success
// return -EINVAL if the dev path has no '/'
// return -ENOENT if this dev has no parent
// return -errno if stat'ing the parent path fails 
// return -ENOMEM if OOM
int sysfs_get_parent_device ( char const* dev_path, char** ret_parent_device, size_t* ret_parent_device_len )
{
   
   char* parent_path = NULL;
   size_t parent_path_len = 0;
   struct stat sb;
   int rc = 0;
   
   char* delim = NULL;
   
   parent_path = (char*)calloc( strlen(dev_path) + 1 + strlen("/uevent"), 1 );
   if( parent_path == NULL ) {
      return -ENOMEM;
   }
   
   strcpy( parent_path, dev_path );
   
   while( strlen(parent_path) > 0 ) {
      
      // lop off the child 
      delim = rindex( parent_path, '/' );
      if( delim == NULL ) {
         
         // invalid 
         free( parent_path );
         return -EINVAL;
      }
      
      if( delim == parent_path ) {
         
         // reached /
         free( parent_path );
         return -ENOENT;
      }
      
      while( *delim == '/' && delim != parent_path ) {
         *delim = '\0';
         delim--;
      }
      
      parent_path_len = strlen( parent_path );
      
      // verify that this is a device...
      strcat( parent_path, "/uevent" );
      
      rc = stat( parent_path, &sb );
      if( rc != 0 ) {
         
         // not a device
         // remove /uevent 
         parent_path[ parent_path_len ] = '\0';
         continue;
      }
      else {
         
         // device!
         break;
      }
   }
   
   // lop off /uevent 
   parent_path[ parent_path_len ] = '\0';
   
   *ret_parent_device = parent_path;
   *ret_parent_device_len = parent_path_len;
   
   return 0;
}

bool sysfs_get_parent_with_subsystem( char const* device_path, char const* subsystem_name )
{
   
   bool res = false;
   char cur_dir[4097];
   char subsystem_path[4097];
   char subsystem_link[4097];
   
   memset( subsystem_path, 0, 4097 );
   memset( subsystem_link, 0, 4097 );
   
   char* tmp = NULL;
   int rc = 0;
   char* parent_device = NULL;
   size_t parent_device_len = 0;
   
   strcpy( cur_dir, device_path );
   
   while(1) {   
      
      // get parent device 
      rc = sysfs_get_parent_device( cur_dir, &parent_device, &parent_device_len );
      if( rc != 0 ) {
         break;
      }
      
      // subsystem?
      sprintf( subsystem_path, "%s/subsystem", parent_device );
      
      memset( subsystem_link, 0, 4096 );
      
      rc = readlink( subsystem_path, subsystem_link, 4096 );
      if( rc < 0 ) {
         
         rc = -errno;
         if( rc != -ENOENT ) {
            fprintf(stderr, "[WARN]: readlink('%s') errno = %d\n", subsystem_path, rc );
         }
         
         free( parent_device );
         parent_device = NULL;
         return res;
      }
      
      // get subsystem name...
      tmp = rindex( subsystem_link, '/' );
      
      if( tmp != NULL ) {
		  if(strcmp( tmp + 1, subsystem_name ) != 0 ) {
         
         // subsystem does not match
         // crawl up to the parent
         strcpy( cur_dir, parent_device );
         
         free( parent_device );
         parent_device = NULL;
         continue;
         }
         else {
		    res = true;
		    break;
	     }
	  }   
	  else
	    break;
  }
  
  return res;
}
