
#include "libvdev/sglib.h"
#include "vdev-modprobe-util.h"

#include <stdbool.h>
#include <sys/wait.h>

#include <poll.h>
#include <linux/netlink.h>
#include <sys/socket.h>
#include <fcntl.h>

#include <sys/time.h>                // for gettimeofday()


#define VDEV_LINUX_NETLINK_BUF_MAX 4097

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

// connection to the linux kernel for hotplug
struct vdev_linux_context {
   
   // netlink address
   struct sockaddr_nl nl_addr;
   
   // poll on the netlink socket
   struct pollfd pfd;
   
   // path to mounted sysfs 
   char sysfs_mountpoint[ PATH_MAX+1 ];
};

struct sysfs_scan_context {
   
   char* uevent_path;
   struct sglib_cstr_vector* frontier;
};


// scan a directory in /sys/devices directory, to find its child directories, that are pushed back to scan_ctx->frontier
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat
static
int vdev_linux_sysfs_scan_device_directory ( char const* fp, void* cls ) {
   
   struct sysfs_scan_context *scan_ctx = (struct sysfs_scan_context*)cls;
   
   struct sglib_cstr_vector* frontier = scan_ctx->frontier;
   
   struct stat sb;
   int rc = 0;
   char* fp_base = NULL;
   char* fp_dup = NULL;
   
   fp_base = rindex( fp, '/' ) + 1;
   
   if( fp_base == NULL ) {
      return 0;
   }
   
   // skip . and .. 
   if( strcmp( fp_base, "." ) == 0 || strcmp( fp_base, ".." ) == 0 ) {
      return 0;
   }
   
   // add directories
   rc = lstat( fp, &sb );
   if( rc != 0 ) {
      
      rc = -errno;
      fprintf(stderr, "lstat('%s') rc = %d\n", fp, rc );
      return rc;
   }
   
   if( !S_ISDIR( sb.st_mode ) && strcmp( fp_base, "uevent" ) != 0 ) {
      
      // not a directory, and not a uevent
      return 0;
   }
   
   fp_dup = vdev_strdup_or_null( fp );
   if( fp_dup == NULL ) {
      
      return -ENOMEM;
   }
   
   if( S_ISDIR( sb.st_mode ) ) {
   
     //vdev_debug("Search '%s'\n", fp_dup );
      
      rc = sglib_cstr_vector_push_back( frontier, fp_dup );
      if( rc != 0 ) {
      
         free( fp_dup );
         return rc;
      }
   }
   else {
      
      // this is a uevent; this directory is a device 
      ///printf("Uevent '%s'\n", fp_dup );
      scan_ctx->uevent_path = fp_dup;
   }
   
   return 0;
}

// get a uevent from a uevent file 
// replace newlines with '\0', making the uevent look like it came from the netlink socket
// (i.e. so it can be parsed by vdev_linux_parse_request)
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat or read
static int vdev_linux_sysfs_read_uevent( char const* fp_uevent, char** ret_uevent_buf, size_t* ret_uevent_len ) {
   
   int rc = 0;
   struct stat sb;
   char* uevent_buf = NULL;
   size_t uevent_buf_len = 0;
   size_t uevent_len = 0;
   
   // get uevent size  
   rc = stat( fp_uevent, &sb );
   if( rc != 0 ) {
      
      rc = -errno;
      
      fprintf(stderr, "stat('%s') rc = %d\n", fp_uevent, rc );
      
      return rc;
   }
   else {
      
      uevent_buf_len = sb.st_size;
   }
   
   // read the uevent
   if( fp_uevent != NULL ) {
      
      uevent_buf = VDEV_CALLOC( char, uevent_buf_len );
      if( uevent_buf == NULL ) {
         
         return -ENOMEM;
      }
      
      rc = modprobe_read_file( fp_uevent, uevent_buf, uevent_buf_len );
      if( rc != 0 ) {
         
         // failed in this 
         fprintf(stderr, "modprobe_read_file('%s') rc = %d\n", fp_uevent, rc );
         free( uevent_buf );
      }
      else {
         
         for( unsigned int i = 0; i < uevent_buf_len; i++ ) {
            
            if( uevent_buf[i] == '\n' ) {
               
               uevent_buf[i] = '\0';
            }
         }
         
         // NOTE: the stat size is an upper-bound.  Find the exact number of bytes.
         for( uevent_len = 0; uevent_len < uevent_buf_len; ) {
            
            if( *(uevent_buf + uevent_len) == '\0' ) {
               break;
            }
            
            uevent_len += strlen( uevent_buf + uevent_len ) + 1;
         }
          
         *ret_uevent_buf = uevent_buf;
         *ret_uevent_len = uevent_len;
      }
   }
   
   return rc;
}

// register a device from sysfs, given the path to its uevent file.
// that is, read the uevent file, generate a device request, and add it to the initial_requests list in the ctx.
// NOTE: it is assumed that fp_uevent--the full path to the uevent file--lives in /sys/devices
// return 0 on success
// return -ENOMEM on OOM
// return negative on error.
static int vdev_linux_sysfs_register_device ( char const* fp_uevent ) {
   
   int rc = 0;
   struct stat sb;
   char* uevent_buf = NULL;
   size_t uevent_buf_len = 0;
   char* full_devpath = NULL;
   char* devpath = NULL;
   char* devname = NULL;
   char* delim = NULL;
   
   // get uevent
   rc = vdev_linux_sysfs_read_uevent( fp_uevent, &uevent_buf, &uevent_buf_len );
   if( rc != 0 ) {
      
      fprintf(stderr, "vdev_linux_sysfs_read_uevent('%s') rc = %d\n", fp_uevent, rc );
      
      free( full_devpath );
      return rc;
   }
   
   if( uevent_buf_len == 0 ) {
      
      free( full_devpath );
      free( uevent_buf );
      return 0;
   }
   
   if(sysfs_get_parent_with_subsystem(fp_uevent, "usb"))
     goto Modprobe;      
   
  
   
    
         char s[1024]={0};
         int r = 0;
         const char *dname = dirname (strdup(fp_uevent));
         strncpy_t(s, sizeof(s), dname, strlen(dname));
         strcat(s, "/boot_vga");
         r = stat( s, &sb );
         if( r != 0 ) {
      
            free( uevent_buf );
            return 0;
         }
   

Modprobe:   
   
   for( unsigned int i = 0; i < uevent_buf_len; ) {

      if (!strncmp(uevent_buf+i, "MODALIAS=", 9)) {
		 
		 FILE *pfin = NULL;
         pid_t pid;
         int wstatus;
         char cmd[1024]={0};
         char *tmp = rindex(uevent_buf + i, '=') + 1;
         
         strncpy_t (cmd, sizeof(cmd), "/sbin/modprobe -b -q -- \"", strlen("/sbin/modprobe -b -q -- \""));
         strcat(cmd, tmp);
         strcat(cmd, "\"");
         
         pfin = epopen(cmd, &pid);
         if(pfin) { 
            fclose(pfin);
            waitpid(pid, &wstatus, 0);
         }

         if (wstatus != 0)
            fprintf(stderr, "Modprobe failed: %s\n", tmp);
      }
           
      i += strlen(uevent_buf + i) + 1;
   }
   
   return rc;
}

static
int vdev_linux_sysfs_find_devices_at_frontier (const char *device_frontier, struct sglib_cstr_vector *uevent_paths)
{
   int rc = 0;
   struct sglib_cstr_vector frontier;
   struct sysfs_scan_context scan_ctx;  

   sglib_cstr_vector_init (&frontier);
   
   memset (&scan_ctx, 0, sizeof(struct sysfs_scan_context));
   
   scan_ctx.frontier = &frontier;

   rc = vdev_load_all (device_frontier, vdev_linux_sysfs_scan_device_directory, &scan_ctx);
   if( rc != 0 ) {
      
      fprintf(stderr, "vdev_load_all('%s') rc = %d\n", dirname, rc );
      sglib_cstr_vector_free (&frontier);      
      return rc;
   }
   
   while( 1 ) {
      
      unsigned long len = sglib_cstr_vector_size (&frontier);
      if (len == 0) break;
   
      char* dir = NULL;
      
      dir = sglib_cstr_vector_at (&frontier, len - 1);
      sglib_cstr_vector_set (&frontier, NULL, len - 1);
      
      sglib_cstr_vector_pop_back (&frontier);
      
      // scan for more devices 
      rc = vdev_load_all (dir, vdev_linux_sysfs_scan_device_directory, &scan_ctx);
      if (rc != 0) {
         
         fprintf(stderr, "vdev_load_all('%s') rc = %d\n", dir, rc);
         free (dir);
         break;
      }
      
      // is one of them a uevent?
      if (scan_ctx.uevent_path) {
         
         // yup--remember it
         char* uevent_path = vdev_strdup_or_null (scan_ctx.uevent_path);
         
         if (!uevent_path) {
            
            free (dir);
            free (scan_ctx.uevent_path);
            scan_ctx.uevent_path = NULL;
            
            rc = -ENOMEM;
            break;
         }

         if(strncmp(uevent_path, "/sys/devices/pci", 16)) {
            
            free (dir);
            break;
         }
         
         rc = sglib_cstr_vector_push_back (uevent_paths, uevent_path);
         if (rc < 0) {
            
            free (uevent_path);
            free (dir);
            free (scan_ctx.uevent_path);
            scan_ctx.uevent_path = NULL;
            
            break;
         }
         
         free (scan_ctx.uevent_path);
         scan_ctx.uevent_path = NULL;
         
         rc = vdev_linux_sysfs_register_device( uevent_path );
         if( rc != 0 ) {
         
            fprintf(stderr, "vdev_linux_sysfs_register_device('%s') rc = %d\n", uevent_path, rc );
         }
      }
      
      free (dir);
   }

Out:
   sglib_cstr_vector_free (&frontier);

   return rc;
}

static
int vdev_linux_sysfs_find_devices( struct vdev_linux_context* ctx, struct sglib_cstr_vector* uevent_paths )
{
   int rc = 0;
   struct sglib_cstr_vector frontier;
   struct sysfs_scan_context scan_ctx;
   
   memset (&scan_ctx, 0, sizeof(struct sysfs_scan_context));
      
   char* devroot = NULL;

   sglib_cstr_vector_init (&frontier);
   
   scan_ctx.frontier = &frontier;
   
   devroot = fullpath( ctx->sysfs_mountpoint, "/devices", NULL );     // devroot <= /sys/devices
   if( devroot == NULL )      
      return -ENOMEM;
      
   rc = vdev_load_all (devroot, vdev_linux_sysfs_scan_device_directory, &scan_ctx);
   if( rc != 0 ) {
      
      fprintf(stderr, "vdev_load_all('%s') rc = %d\n", devroot, rc );      
      free( devroot );      
      sglib_cstr_vector_free( &frontier );      
      return rc;
   }
   
   free (scan_ctx.uevent_path);
   scan_ctx.uevent_path = NULL; 
 
   for (unsigned long i = 0; i < sglib_cstr_vector_size (&frontier); i++)
      vdev_linux_sysfs_find_devices_at_frontier (sglib_cstr_vector_at (&frontier, i), uevent_paths);
   
   sglib_cstr_vector_free( &frontier );
   
   return rc;
}

static
void vdev_linux_context_free (struct vdev_linux_context *ctx)
{
   if( ctx != NULL ) {
      
      if( ctx->pfd.fd >= 0 ) {
         close( ctx->pfd.fd );
         ctx->pfd.fd = -1;
      }
   }
}

int main (int argc, char **argv)
{
   int rc;
   struct sglib_cstr_vector uevent_paths;
   struct vdev_linux_context ctx;
   
   /* netlink */
   ssize_t len = 0;
   char uevent_buf[VDEV_LINUX_NETLINK_BUF_MAX];
   char cbuf[CMSG_SPACE(sizeof(struct ucred))];
   struct cmsghdr *chdr = NULL;
   struct ucred *cred = NULL;
   struct msghdr hdr;
   struct iovec iov;
   struct sockaddr_nl cnls;
   size_t slen = 128 * 1024 * 1024;
   int so_passcred_enable = 1;   
   struct timeval t1, t2;
   
   daemonize();
   
   /*  Coldplug phase: get already existent uevents in the sysfs topology  */
   
   sglib_cstr_vector_init (&uevent_paths);
 
   memset (&ctx, 0, sizeof(struct vdev_linux_context));
   snprintf( ctx.sysfs_mountpoint, PATH_MAX, "/sys" );          
   
   rc = vdev_linux_sysfs_find_devices( &ctx, &uevent_paths );
   
   sglib_cstr_vector_free (&uevent_paths);
   
   
   /*  Netlink phase: wait for new uevent files  */
  
   printf("\n\nWaiting for netlink events...\n");
   printf("NOTE: this is a cancellation point! (wait forever)\n");
   printf("      For example, you can trigger new uvent files in the sysfs by unpluggin your mouse and pluggin it again into the usb port\n");
   
   memset(&hdr, 0, sizeof(struct msghdr));

   // start timer
   gettimeofday(&t1, NULL);

   while (1) {
      
      chdr = NULL;
      cred = NULL;

      ctx.nl_addr.nl_family = AF_NETLINK;
      ctx.nl_addr.nl_pid = getpid();
      ctx.nl_addr.nl_groups = NETLINK_KOBJECT_UEVENT;
      
      ctx.pfd.fd = socket( PF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT );
      if( ctx.pfd.fd < 0 ) {
         
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
      
      ctx.pfd.events = POLLIN;
      
      // big receive buffer, if running as root 
      if( geteuid() == 0 ) {
         rc = setsockopt( ctx.pfd.fd, SOL_SOCKET, SO_RCVBUFFORCE, &slen, sizeof(slen) );
         if( rc < 0 ) {
            
            vdev_linux_context_free (&ctx);
            exit(EXIT_FAILURE);
         }
      }
      
      // check credentials of message--only root should be able talk to us
      rc = setsockopt( ctx.pfd.fd, SOL_SOCKET, SO_PASSCRED, &so_passcred_enable, sizeof(so_passcred_enable) );
      if( rc < 0 ) {
         
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
         
      }
         
      // bind to the address
      rc = bind( ctx.pfd.fd, (struct sockaddr*)&ctx.nl_addr, sizeof(struct sockaddr_nl) );
      if( rc != 0 ) {
         
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
  
      // next event
      rc = poll( &ctx.pfd, 1, 30000 ); 

      // stop timer
      gettimeofday(&t2, NULL);
    
      if ((t2.tv_sec - t1.tv_sec) > 2) {
      
         vdev_linux_context_free (&ctx);
         break;
      }
   
      // get the event 
      iov.iov_base = uevent_buf;
      iov.iov_len = VDEV_LINUX_NETLINK_BUF_MAX;
                
      hdr.msg_iov = &iov;
      hdr.msg_iovlen = 1;
   
      // get control-plane messages
      hdr.msg_control = cbuf;
      hdr.msg_controllen = sizeof(cbuf);
   
      hdr.msg_name = &cnls;
      hdr.msg_namelen = sizeof(cnls);

      // get the event 
      len = recvmsg( ctx.pfd.fd, &hdr, 0 );
      if( len < 0 ) {
      
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
   
      // big enough?
      if( len < 32 || len >= VDEV_LINUX_NETLINK_BUF_MAX ) {
      
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
   
      // control message, for credentials
      chdr = CMSG_FIRSTHDR( &hdr );
      if( chdr == NULL || chdr->cmsg_type != SCM_CREDENTIALS ) {
      
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
   
      // get the credentials
      cred = (struct ucred *)CMSG_DATA(chdr);
   
      // if not root, ignore 
      if( cred->uid != 0 ) {
      
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
   
      // if udev, ignore... they are user space /dev events, and we are 
      // interested in driver core uevents that occur in the kernel space only 
      if( memcmp( uevent_buf, "libudev", 8 ) == 0 ) {
      
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }      
      
      
      // from userspace???
      // kernel messages don't come from userspace 
      if( cnls.nl_pid > 0 ) {

         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      }
   
      if( strstr( uevent_buf, "@/" ) == NULL ) {
      
         vdev_linux_context_free (&ctx);
         exit(EXIT_FAILURE);
      } 
   
      for( unsigned int i = 0; i < len; ) {
         
         if (!strncmp(uevent_buf+i, "MODALIAS=", 9)) {
		 
		    FILE *pfin = NULL;
            pid_t pid;
            int wstatus;
            char cmd[1024]={0};
            char *tmp = rindex(uevent_buf + i, '=') + 1;
         
            strncpy_t (cmd, sizeof(cmd), "/sbin/modprobe -b -q -- \"", strlen("/sbin/modprobe -b -q -- \""));
            strcat(cmd, tmp);
            strcat(cmd, "\"");
            
            printf("%s\n", cmd);
            pfin = epopen(cmd, &pid);
            if(pfin) { 
               fclose(pfin);
               waitpid(pid, &wstatus, 0);
            }

            if (wstatus != 0)
               fprintf(stderr, "Modprobe failed: %s\n", tmp);
         }
         
         i += strlen(uevent_buf + i) + 1;
      }
      
      vdev_linux_context_free (&ctx);
   
   } // while
	
   return 0;
}



